from casadi import *
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import json

# local imports
from utils import plotting

# Fetch configuration
with open ('config.json') as config_file:
    config = json.load(config_file)
#print(config)
try:
    input_sequence_file = config['vdv_model']['input_generation']['storage_file']
    initial_state_values_file = config['vdv_model']['initial_state_values_file']
    dataset_file = config["vdv_model"]["simulation_result_dataset_storage_file"]
except:
    print("Error loading file names from config")

#input_sequence_file = "data/vdv_input_sequence.csv"
#initial_state_values_file  = "data/vdv_initial_states.csv"
#dataset_file = "data/vdv_dataset.csv"

# Not used
output_result_file = "data/vdv_simulation_results.csv"
prod_batch_file = "data/prod/vdv_dataset.csv"

# Hard coded values
#n = 14400 # number of iterations # decided by input sequence
#TF = n*dt*60 # Time horizon
#dt = 1/3600 # step size # Sample once per second

# Fetched from config.json

n = config['vdv_model']['n_iterations']
dt = 1/config['vdv_model']['simulation']['samples_per_hour']

# model parameters
params = {  
    "CA0":  5.1,     
    "T0":104.9,      
   "k10": 1.287e12,  
    "k20": 1.287e12, 
    "k30": 9.043e9,  
    "E1": 9758.3,     
    "E2": 9758.3,     
    "E3": 8560,       
    "dHr1": 4.2,     
    "dHr2": -11,     
    "dHr3": -41.85,  
    "rho": 0.9342,   
    "Cp": 3.01,      
    "kw": 4032,      
    "AR": 0.215,     
    "VR": 10.0,      
    "mK": 5,         
    "CpK": 2
}  


nx = 6 # number of states in plant
nu = 2 # number of input control variables

x = SX.sym('x', nx)  # Concentration of A, B, C, D, temperature in CSTR, T and temp of cooling jacket T_k
p = SX.sym('p', nu)  # Feed rate F_in and Jacket cooling rate, Qk

# reaction rates for A->B, B->C and 2A -> D
k1 = params["k10"]*np.exp(-params["E1"]/(x[4] + 273.15))
k2 = params["k20"]*np.exp(-params["E2"]/(x[4] + 273.15))
k3 = params["k30"]*np.exp(-params["E3"]/(x[4] + 273.15))

r1 = k1*params["VR"]*x[0]
r2 = k2*params["VR"]*x[1]
r3 = k3*params["VR"]*(x[0]**2)


# ODEs
# concentrations
dCA = (-r1 - 2*r3 + p[0]*(params["CA0"]-x[0]))/params["VR"]
dCB = (r1 - r2 - p[0]*x[1])/params["VR"]
dCC = (r2 - p[0]*x[2])/params["VR"]
dCD = (r3 - p[0]*x[3])/params["VR"]

# temperatures
dT = (p[0]*(params["T0"]-x[4])/params["VR"] + (params["kw"]*params["AR"]*(x[5]-x[4]) 
-r1*params["dHr1"] - r2*params["dHr2"] - r3*params["dHr3"])/(params["VR"]*params["rho"]*params["Cp"]))
dTk = (p[1] + params["kw"]*params["AR"]*(x[4] - x[5]))/(params["mK"]*params["CpK"])


# Create ODE set and integrator
xdot = vertcat(dCA, dCB, dCC, dCD, dT, dTk)
ode = {'x':x, 'ode': xdot, 'p': p}
opts = {'tf': dt} # Sets correct step size in integrator
ode_solver = integrator('F', 'cvodes', ode, opts)


# Function for simulation
def simulate_vdv(x0, u):
    states = [x0]
    for k in range(u.shape[0]):
        res = ode_solver(x0=x0, p = u[k])
        x0 = res["xf"]
        states.append(x0)
    return np.concatenate(states, axis = 1)

## Beginning of simulation procedure

# Input sequence for simulation
try: 
    df_u = pd.read_csv(input_sequence_file, index_col=0)
except:
    print('Error reading input sequence file. Defaulting to constant u.')
    # Values for u yielding optimal concentration level of A
    u = [14.19, -1113.5]
    u = np.tile(u, (n,1))
    df_u = pd.DataFrame(data=u, columns = ['Fin', 'Qk'])

u = df_u.values
#print(u)

# Inital state values
if config["vdv_model"]["load_initial_state_values_from_file"]:
    try:
        x0 = np.concatenate(pd.read_csv(initial_state_values_file, index_col=0).to_numpy())
        print("x0 read from file:\n", x0)
    except:
        x0 = config['vdv_model']['initial_state_values']
        print("Could not load initial values from file, x0 defaults to:\n", x0)    
else:    
    x0 = config['vdv_model']['initial_state_values']
    print("x0 initialized to:\n", x0)


x0 = DM(x0)

states = simulate_vdv(x0, u)
TF = dt*u.shape[0]*60 #Time horizon in minutes for visualization
plotting.plot_vdv(states, u, TF, source='Casadi Model')


df_states = pd.DataFrame(data=states.T, columns=['CA', 'CB', 'CC', 'CD', 'T', 'Tk'])
df_last_states = pd.DataFrame(data=states.T[-1].reshape((6,1)).T, columns=['CA0', 'CB0', 'CC0', 'CD0','T0', 'TK0'])


data_set = np.concatenate((states.T[:-1], u, states.T[1:]), axis = 1)

## Uncomment to shuffle data set
#np.random.shuffle(data_set)

df_data_set = pd.DataFrame(data=data_set, 
columns=['CAin', 'CBin', 'CCin', 'CDin', 'Tin', 'Tkin', 'Fin', 'Qk', 'CAout', 'CBout', 'CCout', 'CDout', 'Tout','Tkout'])

# Write results to files
try:
    df_last_states.to_csv(initial_state_values_file) # final state values
    df_states.to_csv(output_result_file) # time series for all states
    df_data_set.to_csv(dataset_file) # data set for training ML model
    #df_data_set.to_csv(prod_batch_file) # same as above, but to seperate file to simulate production setting
except:
    print("Error, could not save simulation results to file")
