
import matplotlib.pyplot as plt
import pandas as pd
import torch
from torch.utils.data import DataLoader
from math import sqrt
import numpy as np
from pathlib import Path
import json

# local imports
from neural_net_class import NeuralNet as Net
from utils import plotting

# We set a fixed seed for repeatability
random_seed = 12345 
torch.manual_seed(random_seed)

## Get path to source directory
src_dir = Path(__file__).parent.parent.absolute()

with open (src_dir / 'config.json') as config_file:
    config = json.load(config_file)

# Load dataset
#df = pd.read_csv(src_dir / 'data/vdv_steady_state_dataset.csv', index_col=0)
try: 
    df = pd.read_csv(src_dir / config["neural_net"]["data_set"], index_col=0)
except:
    print("Unable to load dataset")

# Split data sets
low_index = int(df.shape[0]*0.65)
up_index = int(df.shape[0]*0.95)
test_set = df.iloc[low_index:up_index]

# Define the target and features
INPUT_COLS = config['neural_net']['input_cols']
OUTPUT_COLS = config['neural_net']['output_cols']

# Make a copy of the dataset and remove the test data
train_val_set = df.copy().drop(test_set.index)

# Sample validation data without replacement (10%)
val_set = train_val_set.sample(frac=0.1, replace=False, random_state=random_seed)

# The remaining data is used for training (90%)
train_set = train_val_set.copy().drop(val_set.index)

# Check that the numbers add up
n_points = len(train_set) + len(val_set) + len(test_set)
print("Verify:\nLength of dataset = Length of training set + Length of validation set + length of test set")
print(f'{len(df)} = {len(train_set)} + {len(val_set)} + {len(test_set)} = {n_points}')



def train(
        net: torch.nn.Module,
        train_loader: DataLoader,
        val_loader: DataLoader,
        n_epochs: int,
        lr: float,
        l2_reg: float,
) -> torch.nn.Module:
    """
    Train model using mini-batch SGD
    After each epoch, we evaluate the model on validation data

    :param net: initialized neural network
    :param train_loader: DataLoader containing training set
    :param n_epochs: number of epochs to train
    :param lr: learning rate (default: 0.001)
    :param l2_reg: L2 regularization factor (default: 0)
    :return: torch.nn.Module: trained model.
    """

    # Define loss and optimizer
    criterion = torch.nn.MSELoss(reduction='mean')
    optimizer = torch.optim.Adam(net.parameters(), lr=lr)

    # Train Network
    for epoch in range(n_epochs):
        for inputs, labels in train_loader:
            # Zero the parameter gradients (from last iteration)
            optimizer.zero_grad()

            # Forward propagation
            outputs = net(inputs)
            
            # Compute cost function
            batch_mse = criterion(outputs, labels)
            
            reg_loss = 0
            for param in net.parameters():
                reg_loss += param.pow(2).sum()

            cost = batch_mse + l2_reg * reg_loss

            # Backward propagation to compute gradient
            cost.backward()
            
            # Update parameters using gradient
            optimizer.step()
        
        # Evaluate model on validation data
        mse_val = 0
        for inputs, labels in val_loader:
            mse_val += torch.sum(torch.pow(labels - net(inputs), 2)).item()
        mse_val /= len(val_loader.dataset)
        print(f'Epoch: {epoch + 1}: Val MSE: {mse_val}')
        
    return net




# Get input and output tensors and convert them to torch tensors
x_train = torch.from_numpy(train_set[INPUT_COLS].values).to(torch.float)
y_train = torch.from_numpy(train_set[OUTPUT_COLS].values).to(torch.float)

x_val = torch.from_numpy(val_set[INPUT_COLS].values).to(torch.float)
y_val = torch.from_numpy(val_set[OUTPUT_COLS].values).to(torch.float)

# Create dataset loaders
# Here we specify the batch size and if the data should be shuffled
train_dataset = torch.utils.data.TensorDataset(x_train, y_train)
#train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=10, shuffle=True)
train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=config["neural_net"]["training_batch_size"], shuffle=config["neural_net"]["shuffle_training_data"])


val_dataset = torch.utils.data.TensorDataset(x_val, y_val)
val_loader = torch.utils.data.DataLoader(val_dataset, batch_size=len(val_set), shuffle=False)


# # Construct and initialize the model

# TODO: load previously trained model instead of creating new if wanted.

#layers = [len(INPUT_COLS), 50, 50, len(OUTPUT_COLS)]
layers = [len(INPUT_COLS)] + config["neural_net"]["hidden_layers"] + [len(OUTPUT_COLS)]
net = Net(layers)
print(f'Size of layers: {layers}')
print(f'Number of model parameters: {net.get_num_parameters()}')


# # Train the model
# Hyperparameters
n_epochs = config['neural_net']['n_epochs']
lr = config['neural_net']['learning_rate']
l2_reg = config['neural_net']['l2_regularization']

# Load existing net if specified
if config["neural_net"]["further_train_existing_net"]:
    try:
        net.load(src_dir / config["neural_net"]["prev_net_file"])
        print("Loaded existing net from " + config["neural_net"]["prev_net_file"])
    except:
        print("Unable to load existing net from " + config["neural_net"]["prev_net_file"])


net = train(net, train_loader, val_loader, n_epochs, lr, l2_reg)
net.save(src_dir / config["neural_net"]["trained_net_storage_file"])

# # Load model instead of training

# net.load(src_dir / "neural-nets/trained_nn")

# # Evaluate the model on validation data

# Predict on validation data
pred_val = net(x_val)

# Plot prediction
dt = 1/config["vdv_model"]["simulation"]["samples_per_hour"]
input_sequence_val = val_set[['Fin', 'Qk']].values[1:]

# Limit to 0::100 to avoid data point overload
time = np.linspace(0, (y_val.shape[0]-1)*60*dt, y_val.shape[0])
time = time[0::100]
#time = np.linspace(0, (train_val_set.shape[0]-1)*60*dt, train_val_set.shape[0])
#time = time[val_set.index]
sources = {'Neural Net': pred_val.detach().numpy()[0::100].T, 'Model': y_val.detach().numpy()[0::100].T}
plotting.vdv_plot_states_cmp(sources, time = time, output_cols = OUTPUT_COLS, dt = dt)



# plt.figure(figsize=(16, 9))
# plt.plot(y_val.numpy()[0:100], label='Missing T')
# plt.plot(pred_val.detach().numpy()[0:100], label='Estimated T')
# plt.legend()
# plt.show()

# Compute MSE, MAE and MAPE on validation data
print('Error on validation data')

mse_val = torch.mean(torch.pow(pred_val - y_val, 2))
print(f'MSE: {mse_val.item()}')

mae_val = torch.mean(torch.abs(pred_val - y_val))
print(f'MAE: {mae_val.item()}')

mape_val = 100*torch.mean(torch.abs(torch.div(pred_val - y_val, y_val)))
print(f'MAPE: {mape_val.item()} %')


